@@ -5,34 +5,38 @@ BLE Client for Windows 10 systems, implemented with WinRT.
 Created on 2020-08-19 by hbldh <henrik.blidh@nedomkull.com>
 """
 
-import inspect
-import logging
 import asyncio
-from typing_extensions import Literal, TypedDict
+import logging
+import sys
 import uuid
 import warnings
-from functools import wraps
-from typing import Callable, Any, List, Optional, Sequence, Union
+from typing import Any, Dict, List, Optional, Sequence, Union, cast
 
 import async_timeout
 
+if sys.version_info[:2] < (3, 8):
+    from typing_extensions import Literal, TypedDict
+else:
+    from typing import Literal, TypedDict
+
 from bleak_winrt.windows.devices.bluetooth import (
+    BluetoothAddressType,
+    BluetoothCacheMode,
     BluetoothError,
     BluetoothLEDevice,
-    BluetoothCacheMode,
-    BluetoothAddressType,
 )
 from bleak_winrt.windows.devices.bluetooth.genericattributeprofile import (
     GattCharacteristic,
+    GattCharacteristicProperties,
+    GattClientCharacteristicConfigurationDescriptorValue,
     GattCommunicationStatus,
     GattDescriptor,
     GattDeviceService,
+    GattSession,
     GattSessionStatus,
     GattSessionStatusChangedEventArgs,
+    GattValueChangedEventArgs,
     GattWriteOption,
-    GattCharacteristicProperties,
-    GattClientCharacteristicConfigurationDescriptorValue,
-    GattSession,
 )
 from bleak_winrt.windows.devices.enumeration import (
     DeviceInformation,
@@ -43,17 +47,16 @@ from bleak_winrt.windows.devices.enumeration import (
 from bleak_winrt.windows.foundation import EventRegistrationToken
 from bleak_winrt.windows.storage.streams import Buffer
 
-from bleak.backends.device import BLEDevice
-from bleak.backends.winrt.scanner import BleakScannerWinRT
-from bleak.exc import BleakError, PROTOCOL_ERROR_CODES
-from bleak.backends.client import BaseBleakClient
-
-from bleak.backends.characteristic import BleakGATTCharacteristic
-from bleak.backends.service import BleakGATTServiceCollection
-from bleak.backends.winrt.service import BleakGATTServiceWinRT
-from bleak.backends.winrt.characteristic import BleakGATTCharacteristicWinRT
-from bleak.backends.winrt.descriptor import BleakGATTDescriptorWinRT
-
+from ... import BleakScanner
+from ...exc import PROTOCOL_ERROR_CODES, BleakError, BleakDeviceNotFoundError
+from ..characteristic import BleakGATTCharacteristic
+from ..client import BaseBleakClient, NotifyCallback
+from ..device import BLEDevice
+from ..service import BleakGATTServiceCollection
+from .characteristic import BleakGATTCharacteristicWinRT
+from .descriptor import BleakGATTDescriptorWinRT
+from .scanner import BleakScannerWinRT
+from .service import BleakGATTServiceWinRT
 
 logger = logging.getLogger(__name__)
 
@@ -62,25 +65,28 @@ _ACCESS_DENIED_SERVICES = list(
     for u in ("00001812-0000-1000-8000-00805f9b34fb",)  # Human Interface Device Service
 )
 
-_pairing_statuses = {
-    getattr(DevicePairingResultStatus, v): v
-    for v in dir(DevicePairingResultStatus)
-    if "_" not in v and isinstance(getattr(DevicePairingResultStatus, v), int)
-}
-
-
-_unpairing_statuses = {
-    getattr(DeviceUnpairingResultStatus, v): v
-    for v in dir(DeviceUnpairingResultStatus)
-    if "_" not in v and isinstance(getattr(DeviceUnpairingResultStatus, v), int)
-}
-
 # TODO: we can use this when minimum Python is 3.8
 # class _Result(typing.Protocol):
 #     status: GattCommunicationStatus
 #     protocol_error: typing.Optional[int]
 
 
+def _address_to_int(address: str) -> int:
+    """Converts the Bluetooth device address string to its representing integer
+
+    Args:
+        address (str): Bluetooth device address to convert
+
+    Returns:
+        int: integer representation of the given Bluetooth device address
+    """
+    _address_separators = [":", "-"]
+    for char in _address_separators:
+        address = address.replace(char, "")
+
+    return int(address, base=16)
+
+
 def _ensure_success(result: Any, attr: Optional[str], fail_msg: str) -> Any:
     """
     Ensures that *status* is ``GattCommunicationStatus.SUCCESS``, otherwise
@@ -156,7 +162,7 @@ class BleakClientWinRT(BaseBleakClient):
         self,
         address_or_ble_device: Union[BLEDevice, str],
         *,
-        winrt: WinRTClientArgs = {},
+        winrt: WinRTClientArgs,
         **kwargs,
     ):
         super(BleakClientWinRT, self).__init__(address_or_ble_device, **kwargs)
@@ -166,10 +172,12 @@ class BleakClientWinRT(BaseBleakClient):
             self._device_info = address_or_ble_device.details.adv.bluetooth_address
         else:
             self._device_info = None
-        self._requester = None
+        self._requester: Optional[BluetoothLEDevice] = None
+        self._services_changed_events: List[asyncio.Event] = []
         self._session_active_events: List[asyncio.Event] = []
         self._session_closed_events: List[asyncio.Event] = []
         self._session: GattSession = None
+        self._notification_callbacks: Dict[int, NotifyCallback] = {}
 
         if "address_type" in kwargs:
             warnings.warn(
@@ -182,14 +190,34 @@ class BleakClientWinRT(BaseBleakClient):
         self._use_cached_services = winrt.get("use_cached_services")
         self._address_type = winrt.get("address_type", kwargs.get("address_type"))
 
+        self._session_services_changed_token: Optional[EventRegistrationToken] = None
         self._session_status_changed_token: Optional[EventRegistrationToken] = None
         self._max_pdu_size_changed_token: Optional[EventRegistrationToken] = None
 
     def __str__(self):
-        return "BleakClientWinRT ({0})".format(self.address)
+        return f"{type(self).__name__} ({self.address})"
 
     # Connectivity methods
 
+    async def _create_requester(self, bluetooth_address: int) -> BluetoothLEDevice:
+        args = [
+            bluetooth_address,
+        ]
+        if self._address_type is not None:
+            args.append(
+                BluetoothAddressType.PUBLIC
+                if self._address_type == "public"
+                else BluetoothAddressType.RANDOM
+            )
+        requester = await BluetoothLEDevice.from_bluetooth_address_async(*args)
+
+        # https://github.com/microsoft/Windows-universal-samples/issues/1089#issuecomment-487586755
+        if requester is None:
+            raise BleakDeviceNotFoundError(
+                self.address, f"Device with address {self.address} was not found."
+            )
+        return requester
+
     async def connect(self, **kwargs) -> bool:
         """Connect to the specified GATT server.
 
@@ -203,55 +231,63 @@ class BleakClientWinRT(BaseBleakClient):
         # Try to find the desired device.
         timeout = kwargs.get("timeout", self._timeout)
         if self._device_info is None:
-            device = await BleakScannerWinRT.find_device_by_address(
-                self.address, timeout=timeout
+            device = await BleakScanner.find_device_by_address(
+                self.address, timeout=timeout, backend=BleakScannerWinRT
             )
 
-            if device:
-                self._device_info = device.details.adv.bluetooth_address
-            else:
-                raise BleakError(
-                    "Device with address {0} was not found.".format(self.address)
+            if device is None:
+                raise BleakDeviceNotFoundError(
+                    self.address, f"Device with address {self.address} was not found."
                 )
 
-        logger.debug("Connecting to BLE device @ {0}".format(self.address))
+            self._device_info = device.details.adv.bluetooth_address
 
-        args = [
-            self._device_info,
-        ]
-        if self._address_type is not None:
-            args.append(
-                BluetoothAddressType.PUBLIC
-                if self._address_type == "public"
-                else BluetoothAddressType.RANDOM
-            )
-        self._requester = await BluetoothLEDevice.from_bluetooth_address_async(*args)
+        logger.debug("Connecting to BLE device @ %s", self.address)
 
-        if self._requester is None:
-            # https://github.com/microsoft/Windows-universal-samples/issues/1089#issuecomment-487586755
-            raise BleakError(
-                f"Failed to connect to {self._device_info}. If the device requires pairing, then pair first. If the device uses a random address, it may have changed."
-            )
+        loop = asyncio.get_running_loop()
+
+        self._requester = await self._create_requester(self._device_info)
+
+        def handle_services_changed():
+            if not self._services_changed_events:
+                logger.warn("%s: unhandled services changed event", self.address)
+            else:
+                for event in self._services_changed_events:
+                    event.set()
+
+        def services_changed_handler(sender, args):
+            logger.debug("%s: services changed", self.address)
+            loop.call_soon_threadsafe(handle_services_changed)
+
+        self._services_changed_token = self._requester.add_gatt_services_changed(
+            services_changed_handler
+        )
 
         # Called on disconnect event or on failure to connect.
         def handle_disconnect():
-            if self._session_status_changed_token:
-                self._session.remove_session_status_changed(
-                    self._session_status_changed_token
-                )
-                self._session_status_changed_token = None
-
-            if self._max_pdu_size_changed_token:
-                self._session.remove_max_pdu_size_changed(
-                    self._max_pdu_size_changed_token
-                )
-                self._max_pdu_size_changed_token = None
-
             if self._requester:
+                if self._services_changed_token:
+                    self._requester.remove_gatt_services_changed(
+                        self._services_changed_token
+                    )
+                    self._services_changed_token = None
+
                 self._requester.close()
                 self._requester = None
 
             if self._session:
+                if self._session_status_changed_token:
+                    self._session.remove_session_status_changed(
+                        self._session_status_changed_token
+                    )
+                    self._session_status_changed_token = None
+
+                if self._max_pdu_size_changed_token:
+                    self._session.remove_max_pdu_size_changed(
+                        self._max_pdu_size_changed_token
+                    )
+                    self._max_pdu_size_changed_token = None
+
                 self._session.close()
                 self._session = None
 
@@ -274,8 +310,6 @@ class BleakClientWinRT(BaseBleakClient):
 
                 handle_disconnect()
 
-        loop = asyncio.get_running_loop()
-
         # this is the WinRT event handler will be called on another thread
         def session_status_changed_event_handler(
             sender: GattSession, args: GattSessionStatusChangedEventArgs
@@ -289,7 +323,15 @@ class BleakClientWinRT(BaseBleakClient):
             loop.call_soon_threadsafe(handle_session_status_changed, args)
 
         def max_pdu_size_changed_handler(sender: GattSession, args):
-            logger.debug("max_pdu_size_changed_handler: %d", self._session.max_pdu_size)
+            try:
+                max_pdu_size = sender.max_pdu_size
+            except OSError:
+                # There is a race condition where this event was already
+                # queued when the GattSession object was closed. In that
+                # case, we get a Windows error which we can just ignore.
+                return
+
+            logger.debug("max_pdu_size_changed_handler: %d", max_pdu_size)
 
         # Start a GATT Session to connect
         event = asyncio.Event()
@@ -366,7 +408,9 @@ class BleakClientWinRT(BaseBleakClient):
             self._session_closed_events.append(event)
             try:
                 self._requester.close()
-                async with async_timeout.timeout(10):
+                # sometimes it can take over one minute before Windows decides
+                # to end the GATT session/disconnect the device
+                async with async_timeout.timeout(120):
                     await event.wait()
             finally:
                 self._session_closed_events.remove(event)
@@ -441,17 +485,11 @@ class BleakClientWinRT(BaseBleakClient):
                 DevicePairingResultStatus.PAIRED,
                 DevicePairingResultStatus.ALREADY_PAIRED,
             ):
-                raise BleakError(
-                    "Could not pair with device: {0}: {1}".format(
-                        pairing_result.status,
-                        _pairing_statuses.get(pairing_result.status),
-                    )
-                )
+                raise BleakError(f"Could not pair with device: {pairing_result.status}")
             else:
                 logger.info(
-                    "Paired to device with protection level {0}.".format(
-                        pairing_result.protection_level_used
-                    )
+                    "Paired to device with protection level %d.",
+                    pairing_result.protection_level_used,
                 )
                 return True
         else:
@@ -466,29 +504,26 @@ class BleakClientWinRT(BaseBleakClient):
             Boolean on whether the unparing was successful.
 
         """
-
-        # New local device information object created since the object from the requester isn't updated
-        device_information = await DeviceInformation.create_from_id_async(
-            self._requester.device_information.id
+        device = await self._create_requester(
+            self._device_info
+            if self._device_info is not None
+            else _address_to_int(self.address)
         )
-        if device_information.pairing.is_paired:
-            unpairing_result = await device_information.pairing.unpair_async()
 
+        try:
+            unpairing_result = await device.device_information.pairing.unpair_async()
             if unpairing_result.status not in (
-                DevicePairingResultStatus.PAIRED,
-                DevicePairingResultStatus.ALREADY_PAIRED,
+                DeviceUnpairingResultStatus.UNPAIRED,
+                DeviceUnpairingResultStatus.ALREADY_UNPAIRED,
             ):
                 raise BleakError(
-                    "Could not unpair with device: {0}: {1}".format(
-                        unpairing_result.status,
-                        _unpairing_statuses.get(unpairing_result.status),
-                    )
+                    f"Could not unpair with device: {unpairing_result.status}"
                 )
-            else:
-                logger.info("Unpaired with device.")
-                return True
+            logger.info("Unpaired with device.")
+        finally:
+            device.close()
 
-        return not device_information.pairing.is_paired
+        return True
 
     # GATT services methods
 
@@ -499,73 +534,106 @@ class BleakClientWinRT(BaseBleakClient):
            A :py:class:`bleak.backends.service.BleakGATTServiceCollection` with this device's services tree.
 
         """
+        if not self.is_connected:
+            raise BleakError("Not connected")
+
         # Return the Service Collection.
         if self._services_resolved:
             return self.services
-        else:
-            logger.debug("Get Services...")
 
-            # Each of the get_serv/char/desc_async() methods has two forms, one
-            # with no args and one with a cache_mode argument
-            args = []
+        logger.debug("Get Services...")
 
-            # If the os-specific use_cached_services arg was given when BleakClient
-            # was created, the we use the second form with explicit cache mode.
-            # Otherwise we use the first form with no explicit cache mode which
-            # allows the OS Bluetooth stack to decide what is best.
-            if self._use_cached_services is not None:
-                args.append(
-                    BluetoothCacheMode.CACHED
-                    if self._use_cached_services
-                    else BluetoothCacheMode.UNCACHED
-                )
+        # Each of the get_serv/char/desc_async() methods has two forms, one
+        # with no args and one with a cache_mode argument
+        args = []
 
-            services: Sequence[GattDeviceService] = _ensure_success(
-                await self._requester.get_gatt_services_async(*args),
-                "services",
-                "Could not get GATT services",
+        # If the os-specific use_cached_services arg was given when BleakClient
+        # was created, the we use the second form with explicit cache mode.
+        # Otherwise we use the first form with no explicit cache mode which
+        # allows the OS Bluetooth stack to decide what is best.
+        if self._use_cached_services is not None:
+            args.append(
+                BluetoothCacheMode.CACHED
+                if self._use_cached_services
+                else BluetoothCacheMode.UNCACHED
             )
 
-            for service in services:
-                # Windows returns an ACCESS_DENIED error when trying to enumerate
-                # characteristics of services used by the OS, like the HID service
-                # so we have to exclude those services.
-                if service.uuid in _ACCESS_DENIED_SERVICES:
-                    continue
+        # if we receive a services changed event before get_gatt_services_async()
+        # finishes, we need to call it again with BluetoothCacheMode.UNCACHED
+        # to ensure we have the correct services as described in
+        # https://learn.microsoft.com/en-us/uwp/api/windows.devices.bluetooth.bluetoothledevice.gattserviceschanged
+        while True:
+            services_changed_event = asyncio.Event()
+            services_changed_event_task = asyncio.create_task(
+                services_changed_event.wait()
+            )
+            self._services_changed_events.append(services_changed_event)
+            get_services_task = self._requester.get_gatt_services_async(*args)
 
-                self.services.add_service(BleakGATTServiceWinRT(service))
+            try:
+                await asyncio.wait(
+                    [services_changed_event_task, get_services_task],
+                    return_when=asyncio.FIRST_COMPLETED,
+                )
+            finally:
+                services_changed_event_task.cancel()
+                self._services_changed_events.remove(services_changed_event)
+                get_services_task.cancel()
 
-                characteristics: Sequence[GattCharacteristic] = _ensure_success(
-                    await service.get_characteristics_async(*args),
-                    "characteristics",
-                    f"Could not get GATT characteristics for {service}",
+            if not services_changed_event.is_set():
+                break
+
+            logger.debug(
+                "%s: restarting get services due to services changed event",
+                self.address,
+            )
+            args = [BluetoothCacheMode.UNCACHED]
+
+        services: Sequence[GattDeviceService] = _ensure_success(
+            get_services_task.get_results(),
+            "services",
+            "Could not get GATT services",
+        )
+
+        for service in services:
+            # Windows returns an ACCESS_DENIED error when trying to enumerate
+            # characteristics of services used by the OS, like the HID service
+            # so we have to exclude those services.
+            if service.uuid in _ACCESS_DENIED_SERVICES:
+                continue
+
+            self.services.add_service(BleakGATTServiceWinRT(service))
+
+            characteristics: Sequence[GattCharacteristic] = _ensure_success(
+                await service.get_characteristics_async(*args),
+                "characteristics",
+                f"Could not get GATT characteristics for {service}",
+            )
+
+            for characteristic in characteristics:
+                self.services.add_characteristic(
+                    BleakGATTCharacteristicWinRT(
+                        characteristic, self._session.max_pdu_size - 3
+                    )
                 )
 
-                for characteristic in characteristics:
-                    self.services.add_characteristic(
-                        BleakGATTCharacteristicWinRT(
-                            characteristic, self._session.max_pdu_size - 3
+                descriptors: Sequence[GattDescriptor] = _ensure_success(
+                    await characteristic.get_descriptors_async(*args),
+                    "descriptors",
+                    f"Could not get GATT descriptors for {service}",
+                )
+
+                for descriptor in descriptors:
+                    self.services.add_descriptor(
+                        BleakGATTDescriptorWinRT(
+                            descriptor,
+                            str(characteristic.uuid),
+                            characteristic.attribute_handle,
                         )
                     )
 
-                    descriptors: Sequence[GattDescriptor] = _ensure_success(
-                        await characteristic.get_descriptors_async(*args),
-                        "descriptors",
-                        f"Could not get GATT descriptors for {service}",
-                    )
-
-                    for descriptor in descriptors:
-                        self.services.add_descriptor(
-                            BleakGATTDescriptorWinRT(
-                                descriptor,
-                                str(characteristic.uuid),
-                                characteristic.attribute_handle,
-                            )
-                        )
-
-            logger.info("Services resolved for %s", str(self))
-            self._services_resolved = True
-            return self.services
+        self._services_resolved = True
+        return self.services
 
     # I/O methods
 
@@ -589,6 +657,9 @@ class BleakClientWinRT(BaseBleakClient):
             (bytearray) The read data.
 
         """
+        if not self.is_connected:
+            raise BleakError("Not connected")
+
         use_cached = kwargs.get("use_cached", False)
 
         if not isinstance(char_specifier, BleakGATTCharacteristic):
@@ -596,7 +667,7 @@ class BleakClientWinRT(BaseBleakClient):
         else:
             characteristic = char_specifier
         if not characteristic:
-            raise BleakError("Characteristic {0} was not found!".format(char_specifier))
+            raise BleakError(f"Characteristic {char_specifier} was not found!")
 
         value = bytearray(
             _ensure_success(
@@ -610,7 +681,7 @@ class BleakClientWinRT(BaseBleakClient):
             )
         )
 
-        logger.debug(f"Read Characteristic {characteristic.handle:04X} : {value}")
+        logger.debug("Read Characteristic %04X : %s", characteristic.handle, value)
 
         return value
 
@@ -628,11 +699,14 @@ class BleakClientWinRT(BaseBleakClient):
             (bytearray) The read data.
 
         """
+        if not self.is_connected:
+            raise BleakError("Not connected")
+
         use_cached = kwargs.get("use_cached", False)
 
         descriptor = self.services.get_descriptor(handle)
         if not descriptor:
-            raise BleakError("Descriptor with handle {0} was not found!".format(handle))
+            raise BleakError(f"Descriptor with handle {handle} was not found!")
 
         value = bytearray(
             _ensure_success(
@@ -646,7 +720,7 @@ class BleakClientWinRT(BaseBleakClient):
             )
         )
 
-        logger.debug(f"Read Descriptor {handle:04X} : {value}")
+        logger.debug("Read Descriptor %04X : %s", handle, value)
 
         return value
 
@@ -666,12 +740,15 @@ class BleakClientWinRT(BaseBleakClient):
             response (bool): If write-with-response operation should be done. Defaults to `False`.
 
         """
+        if not self.is_connected:
+            raise BleakError("Not connected")
+
         if not isinstance(char_specifier, BleakGATTCharacteristic):
             characteristic = self.services.get_characteristic(char_specifier)
         else:
             characteristic = char_specifier
         if not characteristic:
-            raise BleakError("Characteristic {} was not found!".format(char_specifier))
+            raise BleakError(f"Characteristic {char_specifier} was not found!")
 
         response = (
             GattWriteOption.WRITE_WITH_RESPONSE
@@ -698,9 +775,12 @@ class BleakClientWinRT(BaseBleakClient):
             data (bytes or bytearray): The data to send.
 
         """
+        if not self.is_connected:
+            raise BleakError("Not connected")
+
         descriptor = self.services.get_descriptor(handle)
         if not descriptor:
-            raise BleakError("Descriptor with handle {0} was not found!".format(handle))
+            raise BleakError(f"Descriptor with handle {handle} was not found!")
 
         buf = Buffer(len(data))
         buf.length = buf.capacity
@@ -712,66 +792,31 @@ class BleakClientWinRT(BaseBleakClient):
             f"Could not write value {data} to descriptor {handle:04X}",
         )
 
-        logger.debug(f"Write Descriptor {handle:04X} : {data}")
+        logger.debug("Write Descriptor %04X : %s", handle, data)
 
     async def start_notify(
         self,
-        char_specifier: Union[BleakGATTCharacteristic, int, str, uuid.UUID],
-        callback: Callable[[int, bytearray], None],
+        characteristic: BleakGATTCharacteristic,
+        callback: NotifyCallback,
         **kwargs,
     ) -> None:
-        """Activate notifications/indications on a characteristic.
-
-        Callbacks must accept two inputs. The first will be a uuid string
-        object and the second will be a bytearray.
-
-        .. code-block:: python
-
-            def callback(sender, data):
-                print(f"{sender}: {data}")
-            client.start_notify(char_uuid, callback)
-
-        Args:
-            char_specifier (BleakGATTCharacteristic, int, str or UUID): The characteristic to activate
-                notifications/indications on a characteristic, specified by either integer handle,
-                UUID or directly by the BleakGATTCharacteristic object representing it.
-            callback (function): The function to be called on notification.
+        """
+        Activate notifications/indications on a characteristic.
 
         Keyword Args:
             force_indicate (bool): If this is set to True, then Bleak will set up a indication request instead of a
                 notification request, given that the characteristic supports notifications as well as indications.
-
         """
-        if inspect.iscoroutinefunction(callback):
-
-            def bleak_callback(s, d):
-                asyncio.ensure_future(callback(s, d))
-
-        else:
-            bleak_callback = callback
-
-        if not isinstance(char_specifier, BleakGATTCharacteristic):
-            characteristic = self.services.get_characteristic(char_specifier)
-        else:
-            characteristic = char_specifier
-        if not characteristic:
-            raise BleakError("Characteristic {0} not found!".format(char_specifier))
-
-        if self._notification_callbacks.get(characteristic.handle):
-            await self.stop_notify(characteristic)
-
-        characteristic_obj = characteristic.obj
+        winrt_char = cast(GattCharacteristic, characteristic.obj)
 
         # If we want to force indicate even when notify is available, also check if the device
         # actually supports indicate as well.
         if not kwargs.get("force_indicate", False) and (
-            characteristic_obj.characteristic_properties
-            & GattCharacteristicProperties.NOTIFY
+            winrt_char.characteristic_properties & GattCharacteristicProperties.NOTIFY
         ):
             cccd = GattClientCharacteristicConfigurationDescriptorValue.NOTIFY
         elif (
-            characteristic_obj.characteristic_properties
-            & GattCharacteristicProperties.INDICATE
+            winrt_char.characteristic_properties & GattCharacteristicProperties.INDICATE
         ):
             cccd = GattClientCharacteristicConfigurationDescriptorValue.INDICATE
         else:
@@ -779,26 +824,33 @@ class BleakClientWinRT(BaseBleakClient):
                 "characteristic does not support notifications or indications"
             )
 
-        fcn = _notification_wrapper(bleak_callback, asyncio.get_running_loop())
-        event_handler_token = characteristic_obj.add_value_changed(fcn)
+        loop = asyncio.get_running_loop()
+
+        def handle_value_changed(
+            sender: GattCharacteristic, args: GattValueChangedEventArgs
+        ):
+            value = bytearray(args.characteristic_value)
+            return loop.call_soon_threadsafe(callback, value)
+
+        event_handler_token = winrt_char.add_value_changed(handle_value_changed)
         self._notification_callbacks[characteristic.handle] = event_handler_token
+
         try:
             _ensure_success(
-                await characteristic_obj.write_client_characteristic_configuration_descriptor_async(
+                await winrt_char.write_client_characteristic_configuration_descriptor_async(
                     cccd
                 ),
                 None,
                 f"Could not start notify on {characteristic.handle:04X}",
             )
         except BaseException:
-
             # This usually happens when a device reports that it supports indicate,
             # but it actually doesn't.
             if characteristic.handle in self._notification_callbacks:
                 event_handler_token = self._notification_callbacks.pop(
                     characteristic.handle
                 )
-                characteristic_obj.remove_value_changed(event_handler_token)
+                winrt_char.remove_value_changed(event_handler_token)
 
             raise
 
@@ -813,12 +865,15 @@ class BleakClientWinRT(BaseBleakClient):
                 directly by the BleakGATTCharacteristic object representing it.
 
         """
+        if not self.is_connected:
+            raise BleakError("Not connected")
+
         if not isinstance(char_specifier, BleakGATTCharacteristic):
             characteristic = self.services.get_characteristic(char_specifier)
         else:
             characteristic = char_specifier
         if not characteristic:
-            raise BleakError("Characteristic {} not found!".format(char_specifier))
+            raise BleakError(f"Characteristic {char_specifier} not found!")
 
         _ensure_success(
             await characteristic.obj.write_client_characteristic_configuration_descriptor_async(
@@ -830,15 +885,3 @@ class BleakClientWinRT(BaseBleakClient):
 
         event_handler_token = self._notification_callbacks.pop(characteristic.handle)
         characteristic.obj.remove_value_changed(event_handler_token)
-
-
-def _notification_wrapper(func: Callable, loop: asyncio.AbstractEventLoop):
-    @wraps(func)
-    def notification_parser(sender: Any, args: Any):
-        # Return only the UUID string representation as sender.
-        # Also do a conversion from System.Bytes[] to bytearray.
-        value = bytearray(args.characteristic_value)
-
-        return loop.call_soon_threadsafe(func, sender.attribute_handle, value)
-
-    return notification_parser
